unit Unit1; 

// Implementierung eines rekursiven mathematischen Parsers in Delphi
// -------------------------------------------------------------------
// Die    function evaluate(t: string): string;    berechnet den
// Wert eines aus Zahlzeichen, den Operatoren + - * / ^ und Klammern
// bestehenden Terms.
// Ein Syntaxtest findet dabei zunächst nicht statt.
// Da Syntaxfehler jedoch (in der Regel) zu Laufzeitfehlern führen,
// können sie über eine try-Umgebung abgefangen werden (siehe dazu
// TForm1.EvaluateExpression, wo dies realisiert ist).
// -------------------------------------------------------------------
// entwickelt im Schuljahr 2012/13 im Rahmen einer Informatik-AG
// am Arnold-Janssen-Gymnasium in St. Wendel
// M.Didas

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs,
  StdCtrls, StrUtils, Math;

type

  { TForm1 }

  TForm1 = class(TForm)
    Edit1: TEdit;
    Memo1: TMemo;
    procedure EvaluateExpression(Sender: TObject);
    procedure Init(Sender: TObject);
    procedure Memo1Change(Sender: TObject);
  private
    { private declarations }
  public
    { public declarations }
  end; 

var
  Form1: TForm1;
  nr : integer;

implementation

{$R *.lfm}

// Teil 1) ----------- Hilfsfunktionen zur Termanalyse ------------------------------------------------------------

function global_bracket(s: string): boolean;             // liefert true, falls "große" Klammer den Term umschließt

var klammer : integer;
    i : integer;
begin
  if AnsiMidStr(s,1,1) <> '(' then result := false
  else begin
    klammer := 0;
    i := 1;
    repeat
      if AnsiMidStr(s,i,1) = '(' then klammer := klammer + 1;
      if AnsiMidStr(s,i,1) = ')' then klammer := klammer - 1;
      i := i+1;
    until (klammer=0) or (i = length(s)+1);

    if (i = length(s)+1) and (klammer =0) then result := true
                                          else result := false;
  end;
end;

function last_free_pos(c: char; s: string): integer;     // findet letztes Vorkommen des Zeichens c im String s außerhalb von Klammern
var pos: integer;
    klammer: integer;

begin

      if AnsiPos(c, s) = 0 then result := 0
      else begin
        pos     := length(s)+1;
        klammer := 0;
        repeat
          pos := pos - 1;
          if pos = 0 then break;
          if AnsiMidStr(s, pos, 1) = '(' then klammer := klammer + 1;
          if AnsiMidStr(s, pos, 1) = ')' then klammer := klammer - 1;
        until ((AnsiMidstr(s, pos, 1) = c) and (klammer = 0));
        result := pos;
      end;
end;

function find_last_add(t: string): integer;              // findet letztes Vorkommen eines + oder -

begin
  result := max(last_free_pos('+',t), last_free_pos('-',t));
end;

function find_last_mul(t: string): integer;              // findet letztes Vorkommen eines * oder /

begin
  result := max(last_free_pos('*',t),last_free_pos('/',t));
end;

function find_last_power(t: string): integer;            // findet letztes Vorkommen eines ^ - Symbols

begin
  result := last_free_pos('^',t);
end;

// Teil 2) ----------- Der eigentliche Parser ---------------------------------------------------------------------

function evaluate(t : string): string;                   // berechnet den Wert eines Terms

var pos:       integer;
    op:        string;
    left_arg:  string;
    right_arg: string;

begin

  // Schritt 1: Entferne (falls vorhanden) eine einschließende globale Klammer,  z.B. aus (8+4-5) wird 8+4-5

  if global_bracket(t) = true then t := evaluate(AnsiMidStr(t,2,length(t)-2));

  // Schritt 2: Werte den letzten Additionsoperator außerhalb von Klammern aus.

  pos := find_last_add(t);

  if pos > 0 then begin
      // ermittle den Operator sowie die beiden Argumente
      op := ansimidstr(t,pos,1);
      if pos = 1 then left_arg := '0'      // Sonderfall: Falls ein Vorzeichen (führendes + oder -) vorliegt,
                                           // setze left_arg = 0
                 else left_arg := evaluate(ansileftstr(t, pos-1));
      right_arg := evaluate(ansirightstr(t, length(t)-pos));
      // führe die Rechnung durch (Fallunterscheidung nach Operator)
      if op = '+' then result := FloatToStr(StrToFloat(left_arg) + StrToFloat(right_arg))
                  else result := FloatToStr(StrToFloat(left_arg) - StrToFloat(right_arg));
      exit;
  end;

  // Schritt 3: Werte den letzten Multiplikationsoperator außerhalb von Klammern aus.

  pos := find_last_mul(t);

  if pos > 0 then begin
      op        := ansimidstr(t,pos,1);
      left_arg  := evaluate(ansileftstr(t, pos-1));
      right_arg := evaluate(ansirightstr(t, length(t)-pos));

      if op = '*' then result := FloatToStr(StrToFloat(left_arg) * StrToFloat(right_arg))
                  else result := FloatToStr(StrToFloat(left_arg) / StrToFloat(right_arg));

      exit;
  end;

  // Schritt 4: Werte den letzten Divisionsoperator außerhalb von Klammern aus.

  pos := find_last_power(t);

  if pos > 0 then begin
      left_arg  := evaluate(ansileftstr(t, pos-1));
      right_arg := evaluate(ansirightstr(t, length(t)-pos));

      result := FloatToStr(Power(StrToFloat(left_arg),StrToFloat(right_arg)));
      exit;
  end;

  // Falls nicht weiter vereinfachbar: Liefere Term selbst wieder zurück.

  result := t;

end;

// Teil 3) ----------- Demoeinbindung: Ein einfacher Taschenrechner -----------------------------------------------

procedure TForm1.EvaluateExpression(Sender: TObject);
var expr : string;
    val  : string;
    test : real;

begin
  expr := Edit1.Text;                       // lese Eingabefeld aus

  if expr <> '' then begin
    Edit1.Clear;
    try
      val := evaluate(expr);                // werte Text aus Eingabefeld aus
      test := StrToFloat(val);              // versuche Resultat in Zahl umzuwandeln
      except val :='error';                 // schlägt einer der Befehle fehl: Fehlermeldung
    end;
    Memo1.Append(Format('%3d:  ',[nr])+expr+' = '+val); // Ausgabe Ergebnis bzw. Fehlermeldung
    nr := nr + 1;
  end;
end;

procedure TForm1.Init(Sender: TObject);
begin
  Edit1.Clear;
  Memo1.Clear;
  nr := 1;
end;

procedure TForm1.Memo1Change(Sender: TObject);
begin

end;


end.

